/*
 *  fileops.c
 *  CocoaDU
 *
 *  Created by Matthijs on Tue Jan 06 1970.
 *  Copyright (c) 2001 __MyCompanyName__. All rights reserved.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <assert.h>
#include <errno.h> 

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>

#include "fileops.h"

typedef struct FileNode FileNode;
typedef struct ProgressSpy ProgressSpy;

/*
static double updater_interval,
static void (*updater)(int n,FileNode **));
*/
static void fillin_fields(
    char *name_buffer,
    FileNode *fn) {
  int ret;
  fn->st=malloc(sizeof(struct stat));

  ret=lstat(name_buffer,fn->st);
  if(ret==-1) {
    fprintf(stderr,"stat(\"%s\"): %s\n",
        name_buffer,strerror(errno));
    fn->flags=FN_STAT_ERROR;
    fn->size=0;
    free(fn->st);
    fn->st=(void*)errno;
    return;
  }
    
  fn->flags=0;
  fn->size=fn->st->st_blocks;
  fn->d_size=fn->st->st_size;
  if(fn->st->st_mode & S_IFDIR)
    fn->flags|=FN_STAT_DIR;
  if(fn->st->st_mode & S_IFLNK) 
    fn->flags|=FN_STAT_SYMLINK;  
  if(fn->st->st_mode & 
      (S_IFCHR | S_IFSOCK | S_IFIFO)) 
    fn->flags|=FN_STAT_OTHER;

  { /* handle ressource fork */
    struct stat st;
    char *end_p;
    for(end_p=name_buffer;*end_p;end_p++);
    strcpy(end_p,"/..namedfork/rsrc");
    ret=lstat(name_buffer,&st);
    if(ret==-1) {
      fn->flags|=FN_STAT_RSRC_ERROR;
      fn->r_size=0;    
    } else {
      fn->r_size=st.st_size;
    }
    *end_p=0;
  }
  
}

#define MAX(a,b) ((a)>(b) ? (a) : (b))

static void explore_dir(
    char *name_buffer,
    FileNode *root_node,
    dev_t root_dev,
    ProgressSpy *spy) {
  int ret;
  unsigned long tot_size;
  int max_depth,n_subdir;
  char *end_p;
  
  /*
  {
    static int n=0;
    if((++n)%500==0)
      printf("-> %s\n",name_buffer);
  }
  */
  
  if(spy) { /* see if caller wants to know where we are */
    assert(pthread_mutex_lock(&spy->mutex)==0);
    if(spy->want_snap) {
      spy->percent=spy->pc_base;
      strcpy(spy->current_name,name_buffer);
      pthread_cond_signal(&spy->cond);
    }
    assert(pthread_mutex_unlock(&spy->mutex)==0);
  }
  
  
  { /* add a slash at the end */
    for(end_p=name_buffer;*end_p;end_p++);
    *end_p++='/';
    *end_p=0;
  }
  
  
  { /* explore directory */
    DIR *dir;
    struct dirent *de;
    FileNode **fn_w=&root_node->children; /* where to write next entry */

    dir=opendir(name_buffer);
    if(!dir) {
      root_node->flags|=FN_OPENDIR_ERROR;
      root_node->children=(void*)errno; 
      return;
    }
    
    tot_size=root_node->size;
    n_subdir=0;
    
    while((de=readdir(dir))) {
      FileNode *fn;
      
      if(!strcmp(de->d_name,".") ||
        !strcmp(de->d_name,"..")) continue;
  
      fn=malloc(sizeof(FileNode)+strlen(de->d_name));
      strcpy(fn->name,de->d_name);
  
      *fn_w=fn;
      fn_w=&fn->next;
      
      strcpy(end_p,de->d_name);
  
      fillin_fields(name_buffer,fn);
  
      if(fn->flags & FN_STAT_ERROR) {
        fn->size=-1;
        goto bail;
      }
      
      if(fn->st->st_dev!=root_dev) {
        fn->flags|=FN_MOUNT_POINT;
      }    
      
      if(fn->flags & FN_STAT_DIR) {
        if(fn->flags & FN_MOUNT_POINT) {
          fn->size=-1;
          fn->depth=0;
        } else {
          n_subdir++;          
        }
      } else {
        tot_size+=fn->size;
        fn->depth=0;
      }
      
    bail:
      *end_p=0;
    }
    *fn_w=NULL;
    ret=closedir(dir);
  }
  
  max_depth=0;
  if(n_subdir) {	/* explore subdirs */
    FileNode *fn;
    double pc_base,pc_fraction;
    if(spy) {
      pc_fraction=spy->pc_tot/(n_subdir+1);
      pc_base=spy->pc_base+pc_fraction; 
    }
      
    for(fn=root_node->children;fn;fn=fn->next) 
      if((fn->flags & FN_STAT_DIR) &&
         !(fn->flags & FN_MOUNT_POINT)) {
        strcpy(end_p,fn->name);
        if(spy) {
          spy->pc_base=pc_base;
          spy->pc_tot=pc_fraction;
          explore_dir(name_buffer,fn,root_dev,spy);
          pc_base+=pc_fraction;
        } else
          explore_dir(name_buffer,fn,root_dev,NULL);         
        tot_size+=fn->size;
        max_depth=MAX(max_depth,fn->depth+1);
      }    
  }
  root_node->size=tot_size;
  root_node->depth=max_depth;
}

/* pthread calling stuff */
typedef struct {
  char *name_buffer;
  FileNode *root_node;
  dev_t root_dev;
  ProgressSpy *spy;
} explore_dir_params;

static void* call_explore_dir(void *arg) {
  explore_dir_params *params=arg;
  params->spy->pc_base=0;
  params->spy->pc_tot=100;
//  printf("spy=%
  explore_dir(
      params->name_buffer,
      params->root_node,
      params->root_dev,
      params->spy);
  pthread_mutex_lock(&params->spy->mutex);
  params->spy->percent=101;
  pthread_mutex_unlock(&params->spy->mutex);
  free(params);
  pthread_exit(NULL);

  return NULL;
}


FileNode *explore_tree(
    char *root,
    ProgressSpy *spy) {
  FileNode *root_node=NULL;
  char *name_buffer=malloc(MAXNAMLEN+1);
  char *name;
  dev_t device;
  
  strcpy(name_buffer,root);

  { 
    char *p;
    for(p=name_buffer;*p;p++);
    if(p==name_buffer) {
      fprintf(stderr,"empty dir name\n");
      goto bail;
    }
    p--;
    while(p>=name_buffer && *p=='/') p--;

    if(p<name_buffer) {
      name=name_buffer; /* ="/" */
      name[1]=0;
    } else {
      p[1]=0;
      while(p>=name_buffer && *p!='/') p--;
      name=*p=='/' ? p+1 : p;
    }
  }  
  
  root_node=malloc(sizeof(FileNode)+strlen(name));
  strcpy(root_node->name,name);
  
  fillin_fields(name_buffer,root_node);
  if(root_node->flags & FN_STAT_ERROR) {
    fprintf(stderr,"stat \"%s\": %s\n",
      name_buffer,strerror((int)root_node->st));
    goto bail;
  }
  if(!(root_node->flags & FN_STAT_DIR)) {
    fprintf(stderr," file \"%s\" is not a directory\n",
      name_buffer);
    goto bail;
  }
  
  device=root_node->st->st_dev;
  
  root_node->next=NULL;
  
  if(spy) {
    pthread_t thr;
    explore_dir_params *params=malloc(sizeof(explore_dir_params));
    params->name_buffer=name_buffer;
    params->root_node=root_node;
    params->root_dev=device;
    params->spy=spy;
    pthread_mutex_init(&spy->mutex,NULL);
    pthread_cond_init(&spy->cond,NULL);
    pthread_create(&thr,NULL,&call_explore_dir,params);
    pthread_detach(thr);
  } else {
    explore_dir(name_buffer,root_node,device,NULL);
    free(name_buffer);
  }
  return root_node;
  
bail:

  if(root_node) free(root_node);
  return NULL;
}

void set_unknown_sizes(FileNode *fn,unsigned long us) {
  

}

void dealloc_tree(FileNode *fn) {

}

