/*
 *  fileops.h
 *  CocoaDU
 *
 *  Created by Matthijs on Tue Jan 06 1970.
 *  Copyright (c) 2001 __MyCompanyName__. All rights reserved.
 *
 */
#ifndef FILEOPS_H_INCLUDED
#define FILEOPS_H_INCLUDED

#include <pthread.h>

struct stat;

#define FN_STAT_ERROR 1
#define FN_STAT_DIR 2
#define FN_STAT_SYMLINK 4
#define FN_STAT_OTHER 8
#define FN_STAT_RSRC_ERROR 16
#define FN_OPENDIR_ERROR 32
#define FN_MOUNT_POINT 64


struct FileNode {
  int flags; /* stat error ? dir ? special file ? */
 
  unsigned long size;  
  /* in 512 byte blocks, including children for directories */
 
  unsigned long d_size; /* data size, in bytes */
  unsigned long r_size; /* ressource size, in bytes */  
  int depth;            /* subdir depth 0=file */

  /* tree fields */
  struct FileNode *next;
  struct FileNode *children;
  
  /* remember stat field */
  struct stat *st;
 
  char name[1];         /* to alloc the struct with 1 malloc */
};


/*
 when the calling thread wants a snapshot, it acquires
 mutex, sets want_snap=1, and waits on cond.
 When cond is signaled, want_snap=0, percent and current_name
 are ready. The end is reached when percent=101. 
*/
struct ProgressSpy {
  pthread_mutex_t mutex;
  int want_snap;
  pthread_cond_t cond;
  
  int cs;
  
  volatile double percent; /* percent for progress bar */
  volatile char current_name[2048]; /* currently explored directory */ 
 
  /* private */
  double pc_base,pc_tot;
};


/* 
  if spy==NULL, the call blocks until FileNode is filled in, 
  else, a new thread is created, and the calling thread can follow the 
  progress as explained above. 
*/
struct FileNode *explore_tree(
  char *root,
  struct ProgressSpy *spy);

void dealloc_tree(struct FileNode *fn);

#endif

